<?
/* --------------------------------------------------------------------
   Copyright (c) 2008-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */
/**
 * stream www
 *
 * HTTP/CGI WWW server
 *
 * @link http://dev.pedemont.com/sonic/
 * @link http://dev.pedemont.com/sonic/downloads/src/lib/streams/sonic.stream.www.php
 * @package sonic_stream
 * @version 0.2.24
 * @abstract web server
*/

define('SONIC_WWW_VERSION', '0.2.28');

//NB, function names in rotate mode have no protection,
//    so dont use generic function names for clash safety

class SONIC_STREAM_www extends SONIC_stream
{
/*
                       sock_in                             www
                      |        |                        |       |
                      |        |                        |       |
                  --> |  udp   | ---------------------> |  udp  |
                  <-> |  tcp   | <--------------------> |  tcp  |
client_ip:client_port |________|                        |       |  ip:port
                       _______                          |       |
                  <-- |  tcp/  | <--------------------- |  tcp/ |
                      |  udp   |                        |  udp  |
   mux_host:mux_port  |        |                        |mux_any|
                      |        |                        [mux_src]
//
                  [ tcp-1 ] <<-------------------------->> [ tcp-1 ]

                  [ udp/  ]                                [       ]
                  [ tcp-1 ] >>-------------------------->> [ udp/  ]    
                                                           [ tcp-1 ]
                  [ udp/  ] <<--------------------------<< [       ]
                  [ tcp-2 ]                                [       ]

*/
  /**
   * HTTP response code
   *
   * eg. 200, 404
   *
   * @var integer
  */
  var $http_response_code = 200;
  /**
   * HTTP header tag key value pair array
   *
   * @see sonic.inc::sonic_header2array() usually an array of this type
   * @var array
  */
  var $http_header_array = array();
  /**
   * HTTP buffer
   *
   * @var string
  */
  var $http_buffer = '';
  /**
   * HTTP buffer already initiated or still untouched
   *
   * @var boolean
  */
  var $http_buffer_idle = true;
//
  /**
   * mux'ed resource socket output var
   *
   * @var resource
  */
  var $sock_out = NULL;
//
  /**
   * run once or cycle
   *
   * enable or disable the internal mini loop daemon 
   *
   * @var boolean
  */
  var $cycle = false;
//
  /**
   * rotate mode flag switch
   *
   * default mode behaviour is threaded
   *
   * @var boolean
  */
  var $rotate_mode = true;

  /**
   * signal termination handler 
   *
   * @param integer $sig posix process signal
  */
  function _sonic_stream_www_handler($sig = NULL)
  { 
    //none of this is really needed, PHP will clean up everything
    //but useful for error messages with server reset in IPC mode
    //
    //in thread mode, sonic_thread_handler() will close
    //socket on last instance/thread anyway
    switch ($sig)
    { 
      case SIGTERM:
      case SIGQUIT:
      case SIGINT:
        // rotate mode OR rotate mode with thread support
        if ($this->rotate_mode)
        { 
          sonic_daemon_thread_shutdown($sig);
          $this->rotate_shutdown = true;
        }
        else
        { 
          $this->sonic_stream_www_stop();
          $this->sonic_stream_buffer();
          exit();
        }
      break;

      default:
        die('www says WTF');
    }
  }

  /**
   * stream main (wrapper)
   *
   * this is really a wrapper function for sonic_stream_www_main(), which
   * iterates the same operation in a loop or once depending on internal
   * config switched flag option
   *
   * @see SONIC_stream::sonic_stream_main()
   * @see SONIC_STREAM_www::sonic_stream_www_main()
   * @param resource $sock_in third party support (library mode) override, resource socket bound and listening for connections
  */
  function sonic_stream_main($sock_in = null)
  {
    if ($sock_in)                       //3rd party compat
      $this->sock = &$sock_in;
    else if ($this->sock)               //sonic daemon inherit preload
    {
      $sock_in = &$this->sock;
    }
    else if ($this->config['socket'])   //despro search
    {
      $sock_in = &$this->config['socket'];
    }
    else
      die('! '.$this->sonic_stream_www_prelog().' main() failed: reason: no socket available'."\n");

    //behaviour of stream usually overwrites root socket with accepted socket
    //so storing copy of root socket now
    $sock_root = $this->sock;
    do
    {
      $this->sonic_stream_buffer(); //threaded buffered output flush

      $this->sock = $sock_root; //copy not address equiv.

      $this->sonic_stream_www_main(); //($this->sock);
 
      if ($this->rotate_shutdown)
        break;
    }
    while ($this->cycle);
  }

  /**
   * stream main
   *
   * @see SONIC_stream::sonic_stream_main()
   * @param resource $sock_in third party support (library mode) override, resource socket bound and listening for connections
  */
  function sonic_stream_www_main($sock_in = null)
  {
    $stream_config = $this->config['stream_config'];

    if (file_exists($stream_config['document_root']) && is_dir($stream_config['document_root']))
    {
      if ($stream_config['document_root'] = realpath($stream_config['document_root']))
      {
        if ($this->verbose > 1)
          echo '= '.$this->sonic_stream_www_prelog().' using document root -> \''.$stream_config['document_root'].'\''."\n";
      }
      else
        die('! '.$this->sonic_stream_www_prelog().' main() failed: reason: unable to resolve absolute \'document_root\' \''.$stream_config['document_root'].'\''."\n");
    }
    else
      die('! '.$this->sonic_stream_www_prelog().' main() failed: reason: invalid \'document_root\' \''.$stream_config['document_root'].'\''."\n");

    //$this->func(&$doesnt_work_ffs)
    list($sock_in, $sock_out) = $this->sonic_stream_www_start($sock_in);

    //quit after accept
    if ($this->rotate_shutdown)
    { 
      $this->sonic_stream_www_stop();
      return;
    }

    //pre-cache vhosts
    $vhosts_config = array();
    $service_vhosts = preg_split('/\s*,\s*/', $stream_config['vhosts']);
    foreach ($service_vhosts as $vhost_lname)
    {
      $vhost = $stream_config[$vhost_lname.'-host'];
      $vhost_config = array();

      if ($vhost)
      {
        //pokemon der ass, if you dont want to catch them all, change this
        foreach ($stream_config as $sc_key => $sc_data)
        {
          if (preg_match('/^'.$vhost_lname.'-(.*)$/', $sc_key, $matches));
          else
            continue;

          $vhost_config["$matches[1]"] = $sc_data;
        }
        $vhost_config['name'] = $vhost; //WARN, this might not be worth doing
        $vhosts_config[$vhost] = $vhost_config;
      }
    }
    unset($service_vhosts, $vhost, $vhost_config);

    //shitty webserver part
    //http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
    //http://www.askapache.com/htaccess/apache-status-code-headers-errordocument.html#status-200

    $method = NULL;

    //NB, socket_accept, socket_read, client send buffer, etc may/will affect timeout
    //TODO, (dont care) use ALARM method if in thread mode for proper timeout?
    if ($stream_config['wait_time'] + 0 > 0)
      $wait = $stream_config['wait_time'] + 0;
    else
      $wait = 8;

    $wait_time = time() + $wait;

    $buf = $_buf = '';
    $buf = $_buf = '';
    do
    {
      if (time() > $wait_time)
        die('! '.$this->sonic_stream_www_prelog().' timeout '.$wait.' seconds'."\n");

      if (false == ($_buf = @socket_read($sock_in, SONIC_WWW_BYTE_SIZE, PHP_BINARY_READ)))
        break;
//      else if ($this->verbose > 0) 
//        echo $buf;

      $buf .= $_buf;

      $_h_buf_a = array();
      $_h_buf_a = sonic_header2array($buf);

      if ($method);
      else
      {
        if (preg_match('/^\s*GET/im', $buf)) { $method = 'get'; }
        else if (preg_match('/^\s*POST/im', $buf)) { $method = 'post'; }
        else
        {
          $out =  '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
		.'<HTML><HEAD>'."\n"
		.'<TITLE>400 Bad Request</TITLE>'."\n"
		.'</HEAD><BODY>'."\n"
		.'<H1>Bad Request</H1>'."\n"
		.'Your browser sent a request that this server could not understand.<P>'."\n"
		.'</BODY></HTML>'."\n";

          if ($this->verbose > 0)
          {
            if (sonic_http_env_mini_engine($header_raw, ($this->stunnel_peername ? 'https' : 'http')));
            else
            { 
              echo '= '.$this->sonic_stream_www_prelog().' [HRC:400] ';
              echo 'ERROR_BAD_HTTP_REQUEST -> '.base64_encode($header_raw)."\n";
              $this->sonic_www_exit(400, $out);
              return; //rotate mode
            }

            echo '= '.$this->sonic_stream_www_prelog()
	        .' [HRC:400] ERROR_BAD_HTTP_REQUEST -> '
		.base64_encode($buf)
		."\n";
          }

          $this->sonic_www_exit(400, $out);
          return; //rotate mode
        }

//        $wait_time += $wait; //timed DDOS?
      }

      if ($method == 'get')
      {
        if (preg_match('/\r{0,1}\n\r{0,1}\n$/', $buf))
        {
//echo "get break\n";
          break;
        }
      }
      else if ($method == 'post')
      {
        if (preg_match('/\r{0,1}\n\r{0,1}\n/', $buf))
        {
  	  //NB, only first header, doesnt handle multiple headers!
          list($_h_buf_h, $_h_buf_b) = preg_split('/\r{0,1}\n\r{0,1}\n/', $buf, 2);
          if ($_h_buf_a['Content-Length'])
          {
            if (($_h_buf_a['Content-Length'] + 0) > strlen($_h_buf_b));
            else
            {
              unset($_h_buf_h, $_h_buf_b);
              break;
            }
          }
          else
            break;
        }
      }
    } while (true);
    unset($_h_buf_a);

    //no point going on if we havent made it this far
    if ($this->rotate_shutdown) // rotate mode
    {
      $this->sonic_stream_www_stop();
      return;
    }

    $header_raw = $buf;
    $header = sonic_header2array($buf);
    unset($buf);

    $_SERVER['SERVER_PROTOCOL'] = 'HTTP/1.1';
    $_SERVER['GATEWAY_INTERFACE'] = 'CGI/1.1';

    //unhook webserver here
//$_SERVER['SCRIPT_FILENAME']
    unset($_GET, $_COOKIE, $_POST, $_FILES, $_REQUEST);
    sonic_http_env_server_var_reset();

    @socket_getsockname($this->sock, $local_ip, $local_port);
    $local_port += 0;
    if ($local_port > 0)
    {
      $_SERVER['SERVER_ADDR'] = $local_ip;
      $_SERVER['SERVER_PORT'] = $local_port;
    }
    else
    {
      $_SERVER['SERVER_ADDR'] = $this->config['ip'];
      $_SERVER['SERVER_PORT'] = $this->config['port'];
    }
    unset($local_port, $local_ip);

    $_she_opts['default_scheme'] = strtolower($stream_config['she_default_scheme']);
    switch ($_she_opts['default_scheme'])
    {
      case 'http':
      case 'https':
      break;

      default:
        $_she_opts['default_scheme'] = ($this->stunnel_peername ? 'https' : 'http');
    }

    if ($header['Host'])
    {
      if (preg_match('/:[0-9]+$/', $header['Host'])) {}
      else if ($_she_opts['default_scheme'] == 'http')
      {
        $header['Host'] .= ':80';
      }
      else //if ($_she_opts['default_scheme'] == 'https')
      {
        $header['Host'] .= ':443';
      }
    }
    else
      $header['Host'] = $_SERVER['SERVER_ADDR'].':'.$_SERVER['SERVER_PORT'];

    $_SERVER['HTTP_HOST'] = $header['Host'];

    if ($vhosts_config["$header[Host]"])
    {
      $_SERVER['DOCUMENT_ROOT'] = $vhosts_config["$header[Host]"]['document_root'];
    }
    else if ($_SERVER['HTTP_HOST'] == $_SERVER['SERVER_ADDR'].':'.$_SERVER['SERVER_PORT'])
    {
      $_SERVER['DOCUMENT_ROOT'] = $stream_config['document_root'];
    }
    else
    {
      //TODO: throw no virtual host error instead of critical?
      unset($_SERVER['DOCUMENT_ROOT']);
    }

    //WARN, is address assign here a bad thing? not sure...
    $_SERVER['REDIRECT_STATUS'] = &$this->http_response_code;

    //s_s_socket_accept() should have init s_s_socket_getpeername already
    //so dont bother calling it again
    if ($this->stunnel_peername)
    {
      list($remote_ip, $remote_port) = preg_split('/:/', $this->stunnel_peername);
    }
    else if ($this->sock_peername)
    {
      list($remote_ip, $remote_port) = preg_split('/:/', $this->sock_peername);
    }
    else
      $this->sock_peername = @socket_getpeername($this->sock, $remote_ip, $remote_port);

    $remote_port += 0;
    if ($remote_port > 0)
    {
      $_SERVER['REMOTE_ADDR'] = $remote_ip;
      $_SERVER['REMOTE_PORT'] = $remote_port;
    }
    unset($remote_port, $remote_ip);

    if (file_exists($_SERVER['DOCUMENT_ROOT']) && is_dir($_SERVER['DOCUMENT_ROOT']));
    else
    {
      $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
		.'<html><head>'."\n"
		.'<title>500 Internal Server Error</title>'."\n"
		.'</head><body>'."\n"
		.'<h1>Internal Server Error</h1>'."\n"
		.'<p>The server encountered a critical internal error or'."\n"
		.'misconfiguration and was unable to complete'."\n"
		.'your request.</p>'."\n"
		.'</body></html>'."\n";

      if ($this->verbose > 0)
      {
        if (sonic_http_env_mini_engine($header_raw, ($this->stunnel_peername ? 'https' : 'http')));
        else
        {
          echo '= '.$this->sonic_stream_www_prelog().' [HRC:500] ';
          echo 'ERROR_BAD_HTTP_REQUEST -> '.base64_encode($header_raw)."\n";
          $this->sonic_www_exit(500, $out);
          return; //rotate mode
        }

        $_dexit = '';

	if ($header['Host'])
          $_dexit .= ':'.$header['Host'];

//	if ($header['User-Agent'])
//          $_dexit .= ':'.$header['User-Agent'];

        if ($_dexit == '');
        else
          $_dexit = '['.substr($_dexit, 1).'] ';

        echo '= '.$this->sonic_stream_www_prelog().' [HRC:500] ';
        echo 'ERROR_DOCUMENT_ROOT -> '
            .$_dexit
            .($_SERVER['DOCUMENT_ROOT'] ?
  	      $_SERVER['DOCUMENT_ROOT']
  	    :
  	      'UNDEFINED'
  	    )
  	   ."\n";
      }

      $this->sonic_www_exit(500, $out);
      return; //rotate mode
    }

    //SHE aka Sonic HTTP Environment, setup
    $stream_config['she_set_globals'] .= ''; //re-cast (string)
    switch (strtolower($stream_config['she_set_globals']))
    {
      case 'off':
      case 'false':
      case 'no':
      case 'none':
        $_she_opts['set_globals'] = false;
      break;

      default:
        $_she_opts['set_globals'] = true;
    }

    if (gettype($stream_config['she_default_index']) == 'string')
      $_she_opts['default_index'] = preg_split('/\s*,\s*/', $stream_config['she_default_index']);
    else
      $_she_opts['default_index'] = array('index.php', 'index.html');

    if (file_exists($stream_config['she_tmp_dir']) && is_dir($stream_config['she_tmp_dir']))
      $_she_opts['tmp_dir'] = $stream_config['she_tmp_dir'];
    else
      $_she_opts['tmp_dir'] = SONIC_TMP_PATH;

    $_she_opts['header_parts'] = &$header;

    if (($_she = sonic_http_env_engine($header_raw, $_she_opts)) > 0)
    {
    }
    else if ($_she != -3 && $_she != -5) //404 "Not Found" handled below
    {
//sonic_http_env()
// 1:   OK
// 0:   not get|post aka unknown method //405 "Method Not Allowed"
//-1:   bad get|post aka invalid syntax //400 "Bad Request"
//-2:   bad post			//417 "Expectation Failed"
//-3:   script filename does not exist  //404 !handled out of this scope
//-4:   bad script aka invalid syntax   //400 "Invalid Syntax"
//-5:   no default index found		//404 !handled out of this scope

      switch ($_she)
      {
        case  0:
          $hrc = 405;
          $_she_msg = 'Method Not Allowed';
        break;
        case -2:
          $hrc = 417;
          $_she_msg = 'Expectation Failed';
        break;
        case -4:
          $hrc = 400;
          $_she_msg = 'Invalid Syntax';
        break;

        case -1:
        default:
          $hrc = 400;
          $_she_msg = 'Bad Request';
      }

//    die("im a shitty webserver, not a doctor jim");
      $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
	.'<HTML><HEAD>'."\n"
	.'<TITLE>400 '.$_she_msg.'</TITLE>'."\n"
	.'</HEAD><BODY>'."\n"
	.'<H1>'.$_she_msg.'</H1>'."\n"
	.'Your browser sent a request that this server could not understand.<P>'."\n"
	.'</BODY></HTML>'."\n";

      if ($this->verbose > 0)
      {
        if (sonic_http_env_mini_engine($header_raw, ($this->stunnel_peername ? 'https' : 'http')));
        else
        { 
          echo '= '.$this->sonic_stream_www_prelog().' [HRC:'.$hrc.'] ';
          echo 'ERROR_BAD_HTTP_REQUEST -> '.base64_encode($header_raw)."\n";
          $this->sonic_www_exit($hrc, $out);
          return; //rotate mode
        }

        echo '= '.$this->sonic_stream_www_prelog()
            .' [HRC:'.$hrc.'] ERROR_SHE -> '.$_she_msg
            .' -> '.base64_encode($header_raw)."\n";
      }

      $this->sonic_www_exit($hrc, $out);
      return; //rotate mode
    }
//    else
//    {
//      //-3 or -5
//    }

    if ($_she_opts['set_globals']);
    else
    {
      $_SERVER['SCRIPT_FILENAME'] = $_she_opts['get_globals']['_SERVER']['SCRIPT_FILENAME'];
      $_SERVER['REQUEST_URI'] = $_she_opts['get_globals']['_SERVER']['REQUEST_URI'];
      $_SERVER['QUERY_STRING']  = $_she_opts['get_globals']['_SERVER']['QUERY_STRING'];
      //might not be set
      $_SERVER['SCRIPT_URI'] = $_she_opts['get_globals']['_SERVER']['SCRIPT_URI'];
    }

    unset($_she, $_she_opts);

    //careful careful set - slow much
    //SERVER_ADMIN            webmaster@*.com //parent
    //SERVER_SIGNATURE        no value        //parent
    //SERVER_SOFTWARE         Apache/2.2.4    //parent
    if ($vhosts_config["$header[Host]"]['server_admin'])
      $_server_admin = $vhosts_config["$header[Host]"]['server_admin'];
    else
      $_server_admin = $stream_config['server_admin'];
    if ($_SERVER['SERVER_ADMIN']) {}
    else if ($_server_admin)
    {
      $_SERVER['SERVER_ADMIN'] = $_server_admin;
    }
    else if (gettype($_server_admin) == 'string') {} //&& $_server_admin == '') {}
    else if (gettype($_server_admin) == 'NULL')
    {
      $_SERVER['SERVER_ADMIN'] = 'webmaster@'.($_SERVER['SERVER_NAME'] ? $_SERVER['SERVER_NAME'] : $this->config['ip']);
    }
    unset($_server_admin);

    if ($vhosts_config["$header[Host]"]['server_signature'])
      $_server_sig = $vhosts_config["$header[Host]"]['server_signature'];
    else
      $_server_sig = $stream_config['server_signature'];
    if ($_SERVER['SERVER_SIGNATURE']) {}
    else if ($_server_sig)
    {
      $_SERVER['SERVER_SIGNATURE'] = $_server_sig;
    }
    else if (gettype($_server_sig) == 'string') {}
    else if (gettype($_server_sig) == 'NULL')
    {
      $_SERVER['SERVER_SIGNATURE'] = 'no value';
    }
    unset($_server_sig);

    if ($vhosts_config["$header[Host]"]['server_software'])
      $_server_soft = $vhosts_config["$header[Host]"]['server_software'];
    else
      $_server_soft = $stream_config['server_software'];
    if ($_SERVER['SERVER_SOFTWARE']) {}
    else if ($server_soft)
    {
      $_SERVER['SERVER_SOFTWARE'] = $_server_soft;
    }
    else if (gettype($_server_soft) == 'string') {} 
    else if (gettype($_server_soft) == 'NULL')
    {
      $_SERVER['SERVER_SOFTWARE'] = 'Sonic WWW/'.SONIC_WWW_VERSION.' ('.SONICD_VERSION.'/'.SONIC_VERSION.') ';
    }
    unset($_server_soft);

    //http://httpd.apache.org/docs/2.0/mod/mod_unique_id.html
    //112-bit (32-bit IP address, 32 bit pid, 32 bit time stamp, 16 bit counter)
    //The actual ordering of the encoding is: time stamp, IP address, pid, TODO/counter
    //TODO: Ummm.. and what iz diz?
    $_SERVER['UNIQUE_ID'] = substr(md5(time().$_SERVER['SERVER_ADDR'].$this->pid.rand()), 0, 19);

//    $hrc = &$this->http_response_code;
    $hrc = 200;

//    if ($this->verbose > 0)
//    {
//      echo '= '.$this->sonic_stream_www_prelog().' [HRC:'.$hrc.'] ';
//
//      if ($_SERVER['SCRIPT_URI'])
//        echo 'URL -> '.$_SERVER['SCRIPT_URI'].($_SERVER['QUERY_STRING'] ? '?'.$_SERVER['QUERY_STRING'] : '')."\n";
//      else
//        echo 'FILE -> '.$_SERVER['SCRIPT_FILENAME']."\n"; //$_SERVER['DOCUMENT_ROOT'].$_SERVER['SCRIPT_URL']."\n";
//    }

//var_dump($_GET, $_POST, $_REQUEST, $_COOKIE, $_SERVER);
//var_dump($_REQUEST, $_SERVER);
//phpinfo();

    // error handler function
    //non threaded protection
    if (function_exists('_sonic_stream_www_main_eh'));
    else
    {
//
    /**
     * PHP HTTP/CGI loadable script error handler
     *
     * @see set_error_handler()
     * @return boolean 
    */
    function _sonic_stream_www_main_eh($errno, $errstr, $errfile, $errline)
    {
      if ($_this = &$_SERVER['SONIC_WWW_SELF']) {}
      else if ($_this = &$_SERVER['SONIC_WWW_COPY']) {}
      else
      {
        $_this = null;
        return false;
      }

      $sonic_www_exit = false;
      if ($errstr == $_this->sonic_www_exit_string())
        $sonic_www_exit = true;

      //only stuff that will get reported to client onwards
      if (
//	  ((error_reporting() | $errno) == E_ALL)	&&
          ((error_reporting() & $errno) == 0)		&&
	  !$sonic_www_exit
	 )
        return false;

      //have to set coz we are one set behind loop catch
      //http://au.php.net/manual/en/function.set-error-handler.php
      //
      //The following error types cannot be handled with a user defined function: 
      //E_ERROR, E_PARSE, E_CORE_ERROR, E_CORE_WARNING, E_COMPILE_ERROR, E_COMPILE_WARNING, 
      //and most of E_STRICT raised in the file where set_error_handler() is called
      //--
      //will have to keep and check logs or summat
      $final_flush = false;
      switch ($errno)
      {
        case E_ERROR: //wont reach here !! 
          $out_eh = '<b>Error</b>';
          $final_flush = true;
        break;

        case E_PARSE: //wont reach here !! 
          $out_eh = '<b>Parse Error</b>';
          $final_flush = true;
        break;

        //wont reach here !! 
        case E_CORE_ERROR: 		//php4
          $out_eh = '<b>Core Error</b>';
          $final_flush = true;
        break;

        //wont reach here !! 
        case E_COMPILE_ERROR: 	//php4
          $out_eh = '<b>Compile Error</b>';
          $final_flush = true;
        break;

        case E_USER_ERROR:	 	//php4
          $out_eh = '<b>User Error</b>';
          $final_flush = true;
        break;

        case E_RECOVERABLE_ERROR:	//php5.2 (error)
          $out_eh = '<b>Recoverable Error</b>';
          $final_flush = true;
        break;

        case E_WARNING:
          $out_eh = '<b>Warning</b>';
        break;

        case E_NOTICE:
          $out_eh = '<b>Notice</b>';
        break;

        //wont reach here !! 
        case E_CORE_WARNING:	//php4
          $out_eh = '<b>Core Warning</b>';
        break;

        //wont reach here !! 
        case E_COMPILE_WARNING: 	//php4
          $out_eh = '<b>Compile Warning</b>';
        break;

        case E_USER_WARNING: 	//php4
          $out_eh = '<b>User Warning</b>';
        break;

        case E_USER_NOTICE:		//php4
          $out_eh = '<b>User Notice</b>';
        break;

        case E_STRICT:		//php5 (notice)
          $out_eh = '<b>Notice</b>';
        break;

        case E_ALL:
          $out_eh = '<b>Critical Error</b>';
        break;

        //NYE
//        case 1066:
//	  $out_eh = '<b>Sonic WWW Psuedo Exit</b>';
//          $final_flush = true;
//        break;

        default:
          $out_eh = '<b>System Error</b>';
      }

      if ($sonic_www_exit)
        $out = '';
      else
        $out = '<br />'//."\n" //sonic log strip 1st only \n$
	   .$out_eh.": $errstr in <b>$errfile</b> on line <b>$errline</b><br />\n"
	   .'';

      if (preg_match('/:\s+Cannot\s+modify\s+header\s+information/i', $out))
        $out = preg_replace('/by\s+\([^\)]*\)\s+in/', 'by', $out);

      if ($final_flush)
      {
        $out = ob_get_contents().$out;
        $_this->sonic_stream_www_ob('end_clean'); //best
      }

      if ($_this->sock)
      {
        //windows $this and &$this warning
        //might not be true copy copy, copy?
        if ($_SERVER['SONIC_WWW_SELF'])
        {
          $_SERVER['SONIC_WWW_SELF']->http_buffer .= $out;
          $_SERVER['SONIC_WWW_COPY']->http_buffer = $_SERVER['SONIC_WWW_SELF']->http_buffer;
        }
        else if ($_SERVER['SONIC_WWW_COPY'])
          $_SERVER['SONIC_WWW_COPY']->http_buffer .= $out;
 
        if ($final_flush) //finished - exit
        {
          $_this->sonic_www_exit($_this->http_response_code, $_this->http_buffer);

//optional return
//          if ($sonic_www_exit)
//            $_this->sonic_www_exit(200, $_this->http_buffer);
//          else
//            $_this->sonic_www_exit(500, $_this->http_buffer);

          return true; //rotate mode
        }
      }
      else
        echo $out;

//      return false;
      return true;
    }
//
    }//rotate mode redeclare protection

    if (file_exists($_SERVER['SCRIPT_FILENAME']) && is_file($_SERVER['SCRIPT_FILENAME']))
    {
      $_fe = explode('.', $_SERVER['SCRIPT_FILENAME']);
      $_fes = sizeof($_fe);

      if ($_fes > 0);
      else
        return;

      $ext = array_pop($_fe);

      if ($_fes == 1)
        $ext = basename($ext);

      unset($_fe, $_fes);

      switch (strtolower($ext))
      {
        case 'php':
        case 'php3':
        case 'php4':
        case 'php5':
        case 'php6':

          //NB, this only needs to be turned on if a custom error handler quiets output all together
          //    surpressing the actual filename that would be shown regardless anyway
//          if ($this->verbose > 0)
//          {
//            echo '= '.$this->sonic_stream_www_prelog().' [HRC:'.$hrc.'] ';
//
//            if ($_SERVER['SCRIPT_URI'])
//             echo 'URL -> '.$_SERVER['SCRIPT_URI'].($_SERVER['QUERY_STRING'] ? '?'.$_SERVER['QUERY_STRING'] : '')."\n";
//            else
//              echo 'FILE -> '.$_SERVER['SCRIPT_FILENAME']."\n"; //$_SERVER['DOCUMENT_ROOT'].$_SERVER['SCRIPT_URL']."\n";
//          }

          //NB, memory management see below
          if ($this->cycle)
          {
            $_mc_i = get_included_files();
            $_mc_i_so = sizeof($_mc_i);
            $_mc_f = get_defined_functions();
            $_mc_f = $_mc_f['user'];
            $_mc_f_so = sizeof($_mc_f);
          }

          $old_error_handler = set_error_handler('_sonic_stream_www_main_eh');
//          $old_error_reporting = error_reporting(0);
          $old_cwd = getcwd();

          ini_set('implicit_flush', false);

//          ob_start(); //good
          $this->sonic_stream_www_ob('start'); //best

          //$this->$_this(init) incase die && for script->header
          $_SERVER['SONIC_WWW_SELF'] = &$this;
          $_SERVER['SONIC_WWW_COPY'] = $this;

          //incase exit to ehandler
          $this->http_response_code = $hrc;

          //31536000(1year) * 28 = 883008000 seconds is roughly what my apache+php setup is doing
          $this->php_header_named_line('Expires', gmdate('D, d M Y G:i:s', time()-883008000).' GMT');
          $this->php_header_named_line('Cache-Control', 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0');
          $this->php_header_named_line('Pragma', 'no-cache');

          //TODO: chdir first? also should check if SF exists?
          chdir(dirname($_SERVER['SCRIPT_FILENAME']));

          //NYE
          //'There is currently no way to process the data with output buffering functions in the shutdown function.'
          //http://www.php.net/manual/en/function.register-shutdown-function.php
	  //TODO?: would have to recode for double hit on exit while no
	  //unregister shutdown function exists
//          register_shutdown_function('_sonic_stream_www_main_eh', 1066, $this->sonic_www_exit_string(), $_SERVER['SCRIPT_FILENAME'], 0);

          include($_SERVER['SCRIPT_FILENAME']);

/*
          $_eval = file_get_contents($_SERVER['SCRIPT_FILENAME']);
          $_eval = str_replace('<?php', '', $_eval);
          $_eval = str_replace('php?>', '', $_eval);
          $_eval = str_replace('<?', '', $_eval);
          $_eval = str_replace('?>', '', $_eval);
          eval($_eval);
          unset($_eval);
*/
//          unset($_this);
          unset($_SERVER['SONIC_WWW_SELF'], $_SERVER['SONIC_WWW_COPY']);
          chdir($old_cwd);

          $out = ob_get_contents();

          $this->sonic_stream_www_ob('end_clean'); //best
//          ob_end_clean(); //good

          ini_set('implicit_flush', true);
//          error_reporting($old_error_reporting);
          restore_error_handler(); //$old_error_handler);

          //NB, will run into FATAL function redeclare errors
          //if we hit the same memory space twice
          //TODO, introduce force mode?
          if ($this->cycle)
          {
            $_mc_i2 = get_included_files();
            $_mc_i2_so = sizeof($_mc);
            $_mc_f2 = get_defined_functions();
            $_mc_f2 = $_mc_f2['user'];
            $_mc_f2_so = sizeof($_mc_f2);

            if ($_mc_i2_so == $_mc_i_so) {}
            else if ($_mc_i2_so > 0 OR $_mc_f2_so > $_mc_f_so)
            {
              if ($_mc_i2[$_mc_i2_so - 1] == $_SERVER['SCRIPT_FILENAME']);
              else
              {
                $this->cycle = false;

                if ($this->verbose > 0)
                  echo '$ '.$this->sonic_stream_www_prelog().' Notice: leaving run-mode \'cycle\' for PHP memory management'."\n";
              }
            }
            unset($_mc_i, $_mc_f_so, $_mc_i2, $_mc_i2_so);
            unset($_mc_f, $_mc_f_so, $_mc_f2, $_mc_f2_so);
          }
        break;

        default:

          $fmtime = filemtime($_SERVER['SCRIPT_FILENAME']);

          $etag = '"'.dechex(fileinode($_SERVER['SCRIPT_FILENAME']))
                 .'-'.dechex(filesize($_SERVER['SCRIPT_FILENAME']))
                 .'-'.dechex($fmtime).'"';

//IPC store
//          $slp = true;
//          $shared_www = sonic_daemon_IPC_session($slp, $this->key);

//          if (gettype($shared_www) == 'array');
//          else
//            $shared_www = array();

//          $check_304 = &$shared_www['304_check']["$_SERVER[REMOTE_ADDR]"];

//          if (gettype($check_304) == 'array');
//          else
//            $check_304 = array();

//          if ($check_304_time = $check_304[$etag])
//          {
            $sane_304 = true;

            if ($header['If-None-Match'] == $etag);
            else
              $sane_304 = false;

            if ($header['If-Modified-Since'])
            {
              $imstime = strtotime($header['If-Modified-Since']);

              if ($imstime == $fmtime);
              else
                $sane_304 = false;

              unset($imstime);
            }

//TODO: find out what to really do with this
            if (preg_match('/(max-age)=([0-9]+)/', $header['Cache-Control'], $matches))
            {
              $max_age = $matches[2]+0;

//              if ($check_304_time + $max_age < time())
              if ($max_age == 0);
              else
                $sane_304 = false;

              unset($matches, $max_age);
            }

            if (preg_match('/no-cache/', $header['Cache-Control']))
              $sane_304 = false;

            if (preg_match('/no-cache/', $header['Pragma']))
              $sane_304 = false;

            if ($sane_304)
              $hrc = 304;
//          }
//          else
//            $check_304[$etag] = time();

          $gmfmdate = gmdate('D, d M Y G:i:s', $fmtime).' GMT';
          unset($fmtime, $sane_304);

//IPC store
//          sonic_daemon_IPC_session($slp, $this->key, $shared_www);

          $this->php_header_named_line('Last-Modified', $gmfmdate);
          $this->php_header_named_line('ETag', $etag);

          if ($hrc == 200)
          {
            if ($out = file_get_contents($_SERVER['SCRIPT_FILENAME']));
            else
            {
              //inhert: file exists already
              if ($_fh = fopen($_SERVER['SCRIPT_FILENAME'], 'r'))
              {
                fclose($_fh);

                $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
			.'<html><head>'."\n"
			.'<title>500 Internal Server Error</title>'."\n"
			.'</head><body>'."\n"
			.'<h1>Internal Server Error</h1>'."\n"
			.'<p>The server encountered an internal error or'."\n"
			.'misconfiguration and was unable to complete'."\n"
			.'your request.</p>'."\n"
			.'<p>Please contact the server administrator,'."\n"
			.' webmaster@'.$_SERVER['SERVER_NAME']
			.' and inform them of the time the error occurred,'."\n"
			.'and anything you might have done that may have'."\n"
			.'caused the error.</p>'."\n"
			.'</body></html>'."\n";

                $hrc = 500;
              }
              else
              {
                $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
		.'<HTML><HEAD>'."\n"
		.'<TITLE>403 Forbidden</TITLE>'."\n"
		.'</HEAD><BODY>'."\n"
		.'<H1>Forbidden</H1>'."\n"
		.'You don\'t have permission to access '.$_SERVER['SCRIPT_URL']
		.' on this server.<P>'."\n"
		.'</BODY></HTML>'."\n";

                $hrc = 403;
              }
            }
//            else
//            {
//              $out = '';
//            }
          }
//        break; //default
      }
    }
    else if (file_exists($_SERVER['SCRIPT_FILENAME']) && is_dir($_SERVER['SCRIPT_FILENAME']))
    {
      $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
	.'<HTML><HEAD>'."\n"
	.'<TITLE>403 Forbidden</TITLE>'."\n"
	.'</HEAD><BODY>'."\n"
	.'<H1>Forbidden</H1>'."\n"
	.'You don\'t have +Index permission to access '.$_SERVER['SCRIPT_URL']
	.' on this server.<P>'."\n"
	.'</BODY></HTML>'."\n";

      $hrc = 403;
    }
    else
    {
      if (sonic_http_env_mini_engine($header_raw, ($this->stunnel_peername ? 'https' : 'http')));
      else
      { 
        echo '= '.$this->sonic_stream_www_prelog().' [HRC:404] ';
        echo 'ERROR_BAD_REQUEST -> '.base64_encode($header_raw)."\n";
        $this->sonic_www_exit(404, $out);
        return; //rotate mode
      }

      $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
	.'<HTML><HEAD>'."\n"
	.'<TITLE>404 Not Found</TITLE>'."\n"
	.'</HEAD><BODY>'."\n"
	.'<H1>Not Found</H1>'."\n"
	.'The requested URL '.$_SERVER['SCRIPT_URL'].' was not found on this server.<P>'."\n"
	.'</BODY></HTML>'."\n";

      $hrc = 404;
    }

//    if ($this->verbose > 0 && $this->cycle)
    if ($this->verbose > 0)
    {
      echo '= '.$this->sonic_stream_www_prelog().' [HRC:'.$hrc.'] ';

      if ($_SERVER['SCRIPT_URI'])
        echo 'URL -> '.$_SERVER['SCRIPT_URI'].($_SERVER['QUERY_STRING'] ? '?'.$_SERVER['QUERY_STRING'] : '')."\n";
      else
        echo 'FILE -> '.$_SERVER['SCRIPT_FILENAME']."\n"; //$_SERVER['DOCUMENT_ROOT'].$_SERVER['SCRIPT_URL']."\n";
    }

    //TODO?: should buffer really be included here

    if (!$this->pcntl_loaded)
    {
      $this->sonic_www_exit($hrc, $this->http_buffer.$out, $sock_out);
      //return; //rotate mode
    }
    else
    {
      $this->sonic_www_echo($hrc, $this->http_buffer.$out, $sock_out);

      if ($this->pid == posix_getpid()) //live and let die (only handle ours)
      {
        //incase we were forked wait for children
        //parent pid of child pid that died;
        if ($this->pid != $this->pid_child) //thread support
        {
          do
          {
            $dead_ppid = pcntl_waitpid($this->pid_child, $status, WNOHANG | WUNTRACED);

            if ($dead_ppid == -1)// OR $dead_ppid == $this->pid)
              break;
          }
          while (true);
        }
        else //rotate with thread support
        {
          do
          {
            $dead_ppid = pcntl_waitpid(-1, $status);//, WNOHANG | WUNTRACED);

            if ($dead_ppid == -1)// OR $dead_ppid == $this->pid)
              break;
          }
          while (true);
        }

        $this->sonic_stream_www_stop();
      }
      else //dont close socket
      {
        $this->sonic_stream_buffer();
        exit();
      }
    }
  }

//PHP FUNCS
//TODO add missing or more PHP funcs
/**
 * PHP session_start() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_session_start()
 * @return boolean
*/
function session_start() { return $this->sonic_www_php_session_start(); }
/**
 * PHP session_start() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_session_start()
 * @return boolean
*/
function php_session_start() { return $this->sonic_www_php_session_start(); }
/**
 * PHP session_start()
 *
 * @see session_start()
 * @return boolean
*/
function sonic_www_php_session_start()
{
  //TODO: check buffer throw own error like this is doing
  $_sesh_start = @session_start();

  if ($_sesh_start);
  else
    return $_sesh_start;

  if ($_sesh_id = session_id())
  {
    $_sesh_name = session_name();
    if ($_COOKIE[$_sesh_name] == $_sesh_id);
    else if (ini_get('session.use_cookies'))
    { 
      $cookie_parts = array();
/*
      if (($expire = ini_get('session.cookie_lifetime') + 0) > 0);
      else
        $expire = time() - (60*60*24*365); //-1 year

      $cookie_parts['expires'] = gmdate("D, d M Y H:i:s \G\M\T", $expire);
*/
      if (($expire = ini_get('session.cookie_lifetime') + 0) > 0)
	//TODO: COOKIE DATE double check this, might be OK
        $cookie_parts['expires'] = gmdate('D, d M Y H:i:s \G\M\T', $expire);

      if ($path = ini_get('session.cookie_path'))
        $cookie_parts['path'] = $path;

      if ($domain = ini_get('session.cookie_domain'))
        $cookie_parts['domain'] = $domain;

      if ($secure = ini_get('session.cookie_secure'))
        $cookie_parts['secure'] = $secure;

      if ($httponly = ini_get('session.cookie_httponly'))
        $cookie_parts['httponly'] = $secure;

      $this->sonic_www_php_set_cookie($_sesh_name, $_sesh_id, $cookie_parts);
    }
  }

  return $_sesh_start;
}

/**
 * PHP setcookie() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_setcookie()
 * @return boolean
*/
function setcookie($name, $value = '', $expire = 0, $path = '', $domain  = '', $secure = false, $httponly = false)
{ 
 return $this->sonic_www_php_setcookie($name, $value, $expire, $path, $domain, $secure, $httponly);
}
/**
 * PHP setcookie() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_setcookie()
 * @return boolean
*/
function php_setcookie($name, $value = '', $expire = 0, $path = '', $domain  = '', $secure = false, $httponly = false)
{
  return $this->sonic_www_php_setcookie($name, $value, $expire, $path, $domain, $secure, $httponly);
}
/**
 * PHP setcookie()
 *
 * @see setcookie()
 * @return boolean
*/
function sonic_www_php_setcookie($cookie_name, $cookie_value = '', $expire = 0, $path = '', $domain  = '', $secure = false, $httponly = false)
{
  if ($expire > 0);
  else
    $expire = time() - (60*60*24*365); //-1 year 

	//TODO: COOKIE DATE double check this, might be OK
  $cookie_parts['expires'] = gmdate('D, d M Y H:i:s \G\M\T', $expire);

  if ($path)
    $cookie_parts['path'] = $path;

  if ($domain)
    $cookie_parts['domain'] = $domain;

  if ($secure)
    $cookie_parts['secure'] = '';

  if ($httponly)
    $cookie_parts['httponly'] = '';

  return $this->sonic_www_php_set_cookie($cookie_name, $cookie_value, $cookie_parts);
}

/**
 * PHP setrawcookie() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_setrawcookie()
 * @return boolean
*/
function setrawcookie($name, $value = '', $expire = 0, $path = '', $domain  = '', $secure = false, $httponly = false)
{ 
 return $this->sonic_www_php_setrawcookie($name, $value, $expire, $path, $domain, $secure, $httponly);
}
/**
 * PHP setrawcookie() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_setrawcookie()
 * @return boolean
*/
function php_setrawcookie($name, $value = '', $expire = 0, $path = '', $domain  = '', $secure = false, $httponly = false)
{
  return $this->sonic_www_php_setrawcookie($name, $value, $expire, $path, $domain, $secure, $httponly);
}
/**
 * PHP setrawcookie()
 *
 * @see setrawcookie()
 * @see SONIC_STREAM_www::sonic_www_php_set_cookie()
 * @return boolean
*/
function sonic_www_php_setrawcookie($cookie_name, $cookie_value = '', $expire = 0, $path = '', $domain  = '', $secure = false, $httponly = false)
{
  if ($expire > 0);
  else
    $expire = time() - (60*60*24*365); //-1 year

	//TODO: COOKIE DATE double check this, might be OK
  $cookie_parts['expires'] = gmdate('D, d M Y H:i:s \G\M\T', $expire);

  if ($path)
    $cookie_parts['path'] = $path;

  if ($domain)
    $cookie_parts['domain'] = $domain;

  if ($secure)
    $cookie_parts['secure'] = '';

  if ($httponly)
    $cookie_parts['httponly'] = '';

  return $this->sonic_www_php_set_cookie($cookie_name, $cookie_value, $cookie_parts, '');
}

//for some fracked reason cant have exit() or /_+exit()/
/**
 * PHP exit() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_exit()
*/
function php_exit($val = null)
{
  return $this->sonic_www_php_exit($val);
}
/**
 * PHP exit()
 *
 * @see exit()
 * @see SONIC_STREAM_www::sonic_www_php_exit()
*/
function sonic_www_php_exit($val = null)
{
  if (gettype($val) == 'NULL' OR (gettype($val) == 'integer' && $val == 0))
  {
    if (function_exists('_sonic_stream_www_main_eh'))
    {
      set_error_handler('_sonic_stream_www_main_eh');

      _sonic_stream_www_main_eh(0, $this->sonic_www_exit_string(), __FILE__, 0);
    }
    else
      trigger_error($this->sonic_www_exit_string(), E_USER_ERROR);
  }
  else
  {
    if (function_exists('_sonic_stream_www_main_eh'))
    {
      set_error_handler('_sonic_stream_www_main_eh');

      if (gettype($val) == 'integer')
      {
        _sonic_stream_www_main_eh($val, $this->sonic_www_exit_string(), __FILE__, 0);
      }
      else if (gettype($val) == 'string')
      {
        _sonic_stream_www_main_eh(0, $val, __FILE__, 0);
      }
      else
        trigger_error($val, E_USER_ERROR); //piss this off? "in sonic.stream.www.php on line XXX"
    }
    else
      trigger_error($val, E_USER_ERROR); //piss this off? "in sonic.stream.www.php on line XXX"
  }


  $this->sonic_www_exit($this->http_response_code, $this->http_buffer.ob_get_clean().$val);

  $this->_sonic_stream_www_handler(SIGQUIT);

//  exit($val);
}

/**
 * PHP header() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_header()
*/
function header($hline, $replace = true, $hrcode = null)
{
  if ($hrcode)
    return $this->sonic_www_php_header($hline, $replace, $hrcode);
  else
    return $this->sonic_www_php_header($hline, $replace);
}
/**
 * PHP header() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_header()
*/ 
function php_header($hline, $replace = true, $hrcode = null)
{
  if ($hrcode)
    return $this->sonic_www_php_header($hline, $replace, $hrcode);
  else
    return $this->sonic_www_php_header($hline, $replace);
}
/**
 * PHP header()
 *
 * @see header()
 * @see SONIC_STREAM_www::sonic_www_php_header_named_line()
*/ 
function sonic_www_php_header($hline, $replace = true, $hrcode = null)
{
  if (preg_match('/^\s*HTTP\/([^\s]*)\s+([0-9]+)\s+([^\r\n]+)\r{0,1}\n{0,1}/i', $hline, $matches))
  {
    $hrcode = $matches[2] + 0;
    $hrcode_msg = $matches[3];

    $this->http_response_code = $hrcode;
  }

  if (preg_match('/\s*([^:]+):\s+([^\r\n]+)\r{0,1}\n{0,1}/', $hline, $matches))
  {
    $h_key = $matches[1];
    $h_val = $matches[2];

    //special headers hack
    switch (strtolower($h_key))
    {
      case 'set-cookie':
        list($name) = preg_split('/=/', $h_val);
        $h_val = preg_replace('/^'.preg_quote($name, '/').'=/i', '=', $h_val);
        $h_key = $h_key.': '.$name;
      break;
    }

/*
    if (array_key_exists($h_key, $this->http_header_array))
    {
      if ($replace)
        $this->http_header_array[$h_key] = $h_val;
      else
        $this->http_header_array[$h_key] .= ", $h_val";
    }
    else
    {
      $this->http_header_array[$h_key] = $h_val;
    }
*/
    $this->sonic_www_php_header_named_line($h_key, $h_val, $replace);

    //WARN, NB, TODO might require more atuo response/status codes
    //will default to 200 via lib
    //Location		HTTP/1.x 302 Found
    //WWW-Authenticate	HTTP/1.x 401 Authorization Required
    switch (strtolower($h_key))
    {
      case 'location':
        $this->http_response_code = 302;
      break;
      case 'www-authenticate':
        $this->http_response_code = 401;
      break;
      case 'proxy-authenticate':
        $this->http_response_code = 407;
      break;
      case 'retry-after':
        $this->http_response_code = 503;
      break;
    }
  }

  return;
}

//PHP WISH
//not an actual replacement of stock func()
/**
 * set_cookie() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_set_cookie()
 * @return boolean
*/
function set_cookie($name, $value = '', $parts = array(), $encode = 'rawurl')
{
  return $this->sonic_www_php_set_cookie($name, $value, $parts, $encode);
}
/**
 * set_cookie() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_set_cookie()
 * @return boolean
*/
function php_set_cookie($name, $value = '', $parts = array(), $encode = 'rawurl')
{
  return $this->sonic_www_php_set_cookie($name, $value, $parts, $encode);
}
/**
 * set_cookie()
 *
 * @param string $name
 * @param string $value
 * @param array $parts
 * @param string $encode
 * @return boolean
*/
function sonic_www_php_set_cookie($name, $value = '', $parts = array(), $encode = 'rawurl')
{
  switch (strtolower($encode))
  {
    case 'url':
      $name = urlencode($name);
      $value = urlencode($value);
    break;
    case 'rawurl':
      $name = rawurlencode($name);
      $value = rawurlencode($value);
    break;
  }

  if ($value == '')
    $value = 'deleted';

  $ch = 'Set-Cookie: '.$name.'='.$value;

  foreach ($parts as $pname => $pvalue)
  {
    if ($pvalue == '')
      $ch .= '; '.$pname;
    else
      $ch .= '; '.$pname.'='.$pvalue;
  }

  return $this->sonic_www_php_header($ch, true); 
}

/**
 * header_named_line() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_header_named_line()
*/
function header_named_line($name, $value = null, $replace = true)
{
  return $this->sonic_www_php_header_named_line($name, $value, $replace);
}
/**
 * header_named_line() (reserved alias)
 *
 * @see SONIC_STREAM_www::sonic_www_php_header_named_line()
*/
function php_header_named_line($name, $value = null, $replace = true)
{
  return $this->sonic_www_php_header_named_line($name, $value, $replace);
}
/**
 * header_named_line()
 *
 * @param string $name
 * @param string $value
 * @param boolean $replace
*/
function sonic_www_php_header_named_line($name, $value = null, $replace = true)
{
  if ($value)
  {
//set
    if (array_key_exists($name, $this->http_header_array))
    {
      if ($replace)
        $this->http_header_array[$name] = $value;
      else
        $this->http_header_array[$name] .= ', '.$value;
    }
    else
    {
      $this->http_header_array[$name] = $value;
    }
  }
  else
  {
//unset
//    if (array_key_exists($name, $this->http_header_array))
      unset($this->http_header_array[$name]);
  }
}

//MAIN/WWW OBJ FUNCS
/**
 * main setup
 *
 * load signal handers and reset internal variables incase of cycle.
 * poll socket_accept listening socket for connected client then return
 *
 * @see SONIC_STREAM_www::sonic_stream_www_main()
 * @see SONIC_STREAM_www::sonic_stream_www_prelog()
 * @param resource third party support (library mode) override, listening resource socket 
 * @return array an array with resource sockets for both I/O respectively
*/
function sonic_stream_www_start($sock_in = NULL)
{
  $this->http_response_code = 200;
  $this->http_header_array = array();
  $this->http_buffer = '';
  $this->http_buffer_idle = true;
//
  $this->sock_out = NULL;
//
  $this->cycle = false;
//
  $this->rotate_mode = true;

  if ($this->pcntl_loaded)
  {
    pcntl_signal(SIGTERM, array(&$this, '_sonic_stream_www_handler'));
    pcntl_signal(SIGQUIT, array(&$this, '_sonic_stream_www_handler'));
    pcntl_signal(SIGINT, array(&$this, '_sonic_stream_www_handler'));
  }

  if ($sock_in)                       //3rd party compat
    $this->sock = &$sock_in;
  else if ($this->sock)               //sonic daemon inherit preload
  { 
    $sock_in = &$this->sock;
  }
  else if ($this->config['socket'])   //despro search
  { 
    $sock_in = &$this->config['socket'];
  }
  else
    die('! '.$this->sonic_stream_www_prelog().' main() failed: reason: no socket available'."\n");

//    var_dump($this->config);

  if (($_bs = 0 + $this->config['byte_size']) > 0) {}
  else { $_bs = 2048; }
  define('SONIC_WWW_BYTE_SIZE', $_bs); unset($_bs);

  //NB, rotate mode OR rotate mode with thread support
  //
  //can be setup before daemon calls main, as daemon
  //might have to setup rotate mode with thread
  //support manually
  $this->rotate_mode = ($this->pid == 0 && $this->pid_child == 0) OR ($this->pid == $this->pid_child);

  $stream_config = $this->config['stream_config'];

  //this is not used here, but for loop main func
  //in this class that calls self here
  //NB, sonic_stream_main()
  switch (strtolower($stream_config['run']))
  {
    case 'once':
    case 'noloop':
    case 'nocycle':
    case 'no-loop':
    case 'no-cycle':
      $this->cycle = false;
    break;

    case 'always':
    case 'loop':
    case 'cycle':
      $this->cycle = true;

      if ($this->rotate_mode && $this->verbose > 0)
        echo '$ '.$this->sonic_stream_www_prelog().' Notice: daemon in \'rotate\' mode, stream in run-mode \'cycle\''."\n";
    break;

    default:
      if ($this->rotate_mode)
        $this->cycle = false;
      else
        $this->cycle = true;
  }

  if ($this->verbose > 1)
  {
    if ($msg = $stream_config['msg_listen'])
      $msg .= "\n";
    else
      $msg = 'listening for connection until close'."\n";

    echo '= '.$this->sonic_stream_www_prelog().' '.$msg;
  }

  if ($this->config['socket_type'] == SOCK_STREAM)
  { 
    $this->sonic_stream_buffer(); //threaded buffer, socket_accept will hang

    if ($_sock_in = $this->sonic_stream_socket_accept($sssa_msg))
    { 
      if ($this->verbose > 0)
        echo '= '.$this->sonic_stream_www_prelog().' '.$sssa_msg."\n";
    }
    else
      die('! '.$this->sonic_stream_www_prelog().' main() failed: reason: '.$sssa_msg."\n");

    //WARN, overwrite root with accept socket
    $this->sock = $_sock_in;
    unset($_sock_in);
    $sock_in = &$this->sock;
  }

  //mux sock out
  if ($mux_host = $stream_config['mux_host'])
  {
    if ($mux_port = $stream_config['mux_port']) $mux_port += 0;
    else
      die('! '.$this->sonic_stream_www_prelog().' main() failed: reason: mux destination port not defined in config'."\n");

    $src_ip = $stream_config['mux_src_ip'];

    //SETUP CONNECTION OUT AND BIND SRC IP
    if ($m_domain = $stream_config['mux_socket_domain'])
    { 
      if ($_def = @constant($m_domain))
        $m_domain = $_def;
    }
    else if ($m_domain = $config['socket_domain'])
    { 
      if ($_def = @constant($m_domain))
        $m_domain = $_def;
    }
    else
      $m_domain = SONIC_SOCK_DOMAIN; //eg. AF_INET
    if ($m_type = $stream_config['mux_socket_type'])
    { 
      if ($_def = @constant($m_type))
        $m_type = $_def;
    }
    else if ($m_type = $config['socket_type'])
    { 
      if ($_def = @constant($m_type))
        $m_type = $_def;
    }
    else
      $m_type = SONIC_SOCK_TYPE; //eg. SOCK_STREAM
    if ($m_protocol = $stream_config['mux_socket_protocol'])
    { 
      if ($_def = @constant($m_protocol))
        $m_protocol = $_def;
    }
    else if ($m_protocol = $config['socket_protocol'])
    { 
      if ($_def = @constant($m_protocol))
        $m_protocol = $_def;
    }
    else
      $m_protocol = SONIC_SOCK_PROTOCOL; //eg. SOL_TCP
    if ($m_options = $stream_config['mux_socket_options'])
    { 
      if ($_def = @constant($m_options))
        $m_options = $_def;
    }
    else if ($m_options = $config['socket_options'])
    { 
      if ($_def = @constant($m_options))
        $m_options = $_def;
    }
    else
      $m_options = SONIC_SOCK_OPTIONS;  //eg. reuseaddr=1,

    unset($_def);

    //string tcp -> SOL_TCP map check
    if (gettype($m_protocol) == 'string')
    { 
      if (($_rsp = getprotobyname($m_protocol)) == -1)
        $m_protocol = SONIC_SOCK_PROTOCOL;
      else
      {
        $m_protocol = $_rsp;
        unset($_rsp);
      }
    }

    if (($sock_out = @socket_create($m_domain, $m_type, $m_protocol)) == false)
      die('! '.$this->sonic_stream_www_prelog().' socket_create() failed: reason: '.socket_strerror(socket_last_error())."\n");

    sonic_socket_set_options($sock_out, SOL_SOCKET, $m_options);

    if ($mux_src_ip)
      if (@socket_bind($sock_out, $mux_src_ip) == false)
        die('! '.$this->sonic_stream_www_prelog().' socket_bind() failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n");

    if ($m_type == SOCK_STREAM)
    { 
      if (@socket_connect($sock_out, $mux_host, $mux_port) == false)
        die('! '.$this->sonic_stream_www_prelog().' socket_connect() failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n");
    }
  }

  if ($sock_out)
  {
    $this->sock_out = &$sock_out;
  }
  else if ($this->sock_out)
  {
    $sock_out = &$this->sock_out;
  }
  else
  {
    $this->sock_out = NULL;
    $sock_out = &$sock_in;
  }
  //mux

  return array(&$sock_in, &$sock_out); //ffs
}
/**
 * log output unifier
 *
 * returns a uniform prefix for log/output entries
 *
 * @see SONIC_stream::sonic_stream_prelog()
 * @return string string usually with pid/process id and timestamp
*/
function sonic_stream_www_prelog()
{
  return $this->sonic_stream_prelog();
}
/**
 * main shutdown/finish
 *
 * closes in and out resource socket variables and echos trace
 *
 * @see SONIC_STREAM_www::sonic_stream_www_main()
 * @see SONIC_STREAM_www::sonic_stream_www_prelog()
 * @param string default close exit message override
*/
function sonic_stream_www_stop($msg_close = null)
{
  if ($this->sock)
    socket_close($this->sock);

  if ($this->sock_out)
    socket_close($this->sock_out);

  if ($this->verbose > 1)
  {
    $stream_config = $this->config['stream_config'];

    if ($msg_close)
    {
      $msg_close = strip_tags($msg_close);
      $msg_close = trim($msg_close);

      //first line only
      list($msg_close,) =  preg_split('/\r{0,1}\n/', $msg_close);

      $msg_close .= "\n"; 
    }
    else if ($msg = $stream_config['msg_close'])
    {
      $msg .= "\n";
    }
    else
      $msg_close = 'connection finished'."\n";

    echo '= '.$this->sonic_stream_www_prelog().' [HRC:'.$this->http_response_code.'] '.$msg_close;
  }

//  $this->idle = true;
}

/**
 * ob wrapper
 *
 * @see ob_start()
 * @see ob_end_clean()
 * @return mixed mixed on success false on failure
*/
function sonic_stream_www_ob($ob_func = null)
{
  switch (strtolower($ob_func))
  {
    case 'start':
    case 'end_clean':
      break;

    default:
      return false;
  }

  return call_user_func('ob_'.$ob_func);
}

//MAIN//OBJ//DMZ internal

//exit handler can match this to fane proper exit
/**
 * error string
 * 
 * @see set_error_handler()
 * @return string a unique string used to trigger, catch and fane errors
*/
function sonic_www_exit_string()
{
  //return the _same_ unique string
  //does it really need hashing?
  return 'Sonic WWW pseudo exit error';
}
/**
 * HTTP response with exit
 *
 * wrapper for sonic_www_echo() with exit()
 *
 * @see SONIC_STREAM_www::sonic_www_echo()
 * @param integer $http_response_code HTTP response code
 * @param mixed $out HTTP body data
 * @param resource $sock third party support (library mode) override, output resource socket to write response to
*/
function sonic_www_exit($http_response_code = null, $out = 0, $sock = null)// = $this->sock)
{
  if ($http_response_code);
  else
  {
    if ($out)
      $http_response_code = 500;
    else
      $http_response_code = 200;
  }

  if ($this->verbose > 0) // && $http_response_code != 200)
  { 
    echo '= '.$this->sonic_stream_www_prelog().' [HRC:'.$http_response_code.'] ';

    if ($_SERVER['SCRIPT_URI'])
    {
      echo 'URL -> '.$_SERVER['SCRIPT_URI'].($_SERVER['QUERY_STRING'] ? '?'.$_SERVER['QUERY_STRING'] : '')."\n";
    }
    else if ($_SERVER['SCRIPT_FILENAME'])
    {
      echo 'FILE -> '.$_SERVER['SCRIPT_FILENAME']."\n"; //$_SERVER['DOCUMENT_ROOT'].$_SERVER['SCRIPT_URL']."\n";
    }
    else if ($_SERVER['SCRIPT_URL'] && $_SERVER['DOCUMENT_ROOT'])
    {
      echo 'FILE -> '.$_SERVER['DOCUMENT_ROOT'].$_SERVER['SCRIPT_URL']."\n";
    }
    else if ($_SERVER['DOCUMENT_ROOT'])
    {
      echo 'DIR -> '.$_SERVER['DOCUMENT_ROOT']."\n";
    }
    else
    {
      echo 'ERROR -> BAD/LOST/UNKNOWN REQUEST'."\n";
    }
  }

  if (gettype($out) == 'integer')
  {
    $this->http_response_code = $http_response_code;
  }
  else if (gettype($out) == 'string')
  {
    $this->sonic_www_echo($http_response_code, $out, $sock);
  }
  else
  {
    $out = var_export($out, true);
    $this->sonic_www_echo($http_response_code, $out, $sock);
  }

//  $this->sonic_stream_buffer();
//  exit($out);

  $this->sonic_stream_www_stop();
  if ($this->rotate_mode OR $this->cycle);
  else
  {
    $this->sonic_stream_buffer();
    if ($http_response_code > 399)
      exit(1);
    else
      exit(0);
  }
}

/**
 * HTTP response
 *
 * form and write out HTTP response on internal resource
 * socket or supplied override. given data in $out will be
 * merged with internal http buffer
 *
 * @see SONIC_STREAM_www::$http_header_array
 * @see SONIC_STREAM_www::$http_buffer
 * @see SONIC_STREAM_www::sonic_www_mimetype()
 * @see sonic.inc::sonic_http_response()
 * @param integer $http_response_code HTTP response code
 * @param string $out HTTP body data
 * @param resource $sock third party support (library mode) override, output resource socket to write response to
*/
function sonic_www_echo($http_response_code, $out, $sock = null)
{
//  $this->http_buffer_idle = false;

  if ($sock) {}
  else if ($this->sock_out)
  {
    $sock = &$this->sock_out;
  }
  else if ($this->sock)
  {
    $sock = &$this->sock;
  }
  else
    return false;

  $http_response_code += 0;
  if ($http_response_code > 0);
  else
  {
    $http_response_code = 200;
  }

  $this->http_response_code = $http_response_code;

//echo $out;

  if (file_exists($_SERVER['SCRIPT_FILENAME']) && is_file($_SERVER['SCRIPT_FILENAME']))
  {
    if ($_mime_type = $this->sonic_www_mimetype($_SERVER['SCRIPT_FILENAME']));
    else
      $_mime_type = 'text/html';
  }
  else
    $_mime_type = 'text/html';
//    $_mime_type = 'text/plain';

  $out_headers = $this->http_header_array;

//  if (gettype($out_headers) == 'array');
//  else
//    $out_headers = array();
//
  $out_headers = sonic_http_header_check($out_headers);

  if ($out_headers['Content-Type']);
  else
    $out_headers['Content-Type'] = $_mime_type;

  if ($this->http_buffer_idle)
  {
    $this->http_buffer_idle = false;
    $out_final = sonic_http_response($http_response_code, $out, $out_headers);
  }
  else
    $out_final = &$out;

  $out_final_strlen = strlen($out_final);

  //JIC sock past isnt ours etc
  if (socket_get_option($sock, SOL_SOCKET, SO_TYPE) == SOCK_STREAM)
  {
//TODO: make sonic_socket_write (lock) optional to save filesystem
//    $sock_ret_bytes = sonic_socket_write($sock, $out_final, $out_final_strlen);
    $sock_ret_bytes = @socket_write($sock, $out_final, $out_final_strlen);
  }
  else 
  {
    if ($host = $this->config['stream_config']['mux_host'])
      $port = $this->config['stream_config']['mux_port'] + 0;
    else if ($host = $this->config['ip'])
      $port = $this->config['port'] + 0;

    $sock_ret_bytes = @socket_sendto($sock, $out_final, $out_final_strlen, 0, $host, $port);
  }

  if ($sock_ret_bytes < 0 OR $sock_ret_bytes == false)
    echo '! '.$this->sonic_stream_www_prelog().' echo() failed: reason: '.socket_strerror(socket_last_error($sock))."\n";
}

/**
 * mime type
 *
 * fetch mime type using mime type definition file
 *
 * @return string string on success or false on failure
*/
function sonic_www_mimetype($file, $mimetype_file = SONIC_MIME_FILE)
{
//  $mimetype_file = '/etc/mime.types';

  $_fe = explode('.', $file);
  $_fes = sizeof($_fe);

  if ($_fes > 0);
  else
    return;

  $ext = array_pop($_fe);

  if ($_fes == 1)
   $ext = basename($ext);

  unset($_fe, $_fes);

  switch (strtolower($ext))
  {
    case 'php':
    case 'php3':
    case 'php4':
    case 'php5':
      return 'text/html';
  }

  $pext = str_replace(DIRECTOARY_SEPARATOR, '\\'.DIRECTORY_SEPARATOR, $ext);
  $mfi = file($mimetype_file);
  foreach ($mfi as $line)
  {
    //NB, grabs commented
    if (preg_match('/\s+'.$pext.'\s+/', $line))
      if (preg_match('/^\s*[\s#]*([^\s]+)\s+/', $line, $m))
        return $m[1];
  }

  if (function_exists('mime_content_type'));
  else
  {
    function mime_content_type($file)
    {
      $fi = finfo_open(FILEINFO_MIME);
      $mimetype = finfo_file($fi, $file);
      finfo_close($fi);
      return $mimetype;
    }
  }

  if (file_exists($file));
  else
    return;

  return mime_content_type($file);
//  return 'application/octet-stream';
}

//thats a wrap class
}

?>
